#include <sys/types.h>
#include <sys/file.h>
#include <a.out.h>
#include <stdio.h>
#include <sys/conf.h>
#include <sys/param.h>
#include <machine/pte.h>
#include <sys/time.h>
#include <sys/proc.h>

extern int errno;

static int pmemfd = -1;
static caddr_t pmemnextaddr = 0;

/*
 * pfetch(ADDR) - fetch one word (32 bits) at ADDR from physical memory
 *
 */

long
pfetch(addr)
caddr_t addr;
{
  daddr_t buf;
  int cc;

  if (pmemfd == -1) {
    pmemfd = open("/dev/mem", O_RDONLY, 0);
    if (pmemfd < 0) {printf("Error opening /dev/mem, errno=%d\n", errno);
		 exit(1);
	       }
  }

  if (addr != pmemnextaddr)
    if (lseek(pmemfd, addr-pmemnextaddr, L_INCR) == -1) {
      printf("Physical lseek failed, errno=%d\n", errno);
      exit(1);
    }

  cc = read(pmemfd, &buf, sizeof(buf));

  if (cc != sizeof(buf)) {
    printf("Physical read failed, errno=%d\n", errno);
    exit(1);
  }

  pmemnextaddr = addr + sizeof(daddr_t);
  return(buf);
}

/*
 * pfetchm(KERN_ADDR, MY_ADDR, LENGTH) - fetch LENGTH bytes at ADDR in
 * physical memory to MY_ADDR.
 *
 */

pfetchm(kern_addr, my_addr, length)
caddr_t kern_addr;
unsigned char *my_addr;
int length;
{
  int cc;

  if (pmemfd == -1) {
    pmemfd = open("/dev/mem", O_RDONLY, 0);
    if (pmemfd < 0) {printf("Error opening /dev/mem, errno=%d\n", errno);
		 exit(1);
	       }
  }

  if (kern_addr != pmemnextaddr)
    if (lseek(pmemfd, kern_addr, L_SET) == -1) {
      printf("Physical lseek failed, errno=%d\n", errno);
      exit(1);
    }

  cc = read(pmemfd, my_addr, length);

  if (cc != length) {
    printf("Physical read failed, errno=%d\n", errno);
    exit(1);
  }

  pmemnextaddr = kern_addr + length;

}

static int kmemfd = -1;
static daddr_t kmemnextaddr = 0;

/*
 * fetch(ADDR) - Kernal virtual address space fetch
 *
 */

unsigned long
fetch(addr)
daddr_t addr;
{
  daddr_t buf;
  int cc;

  if (kmemfd == -1) {
    kmemfd = open("/dev/kmem", O_RDONLY, 0);
    if (kmemfd < 0) {printf("Error opening /dev/kmem, errno=%d\n", errno);
		 exit(1);
	       }
  }

  if (addr != kmemnextaddr)
    if (lseek(kmemfd, addr, L_SET) == -1) {
      printf("Virtual lseek failed, errno=%d\n", errno);
      exit(1);
    }

  cc = read(kmemfd, &buf, sizeof(buf));

  if (cc != sizeof(buf)) {
    printf("Virtual read failed, errno=%d\n", errno);
    exit(1);
  }

  kmemnextaddr = addr + sizeof(daddr_t);
  return(buf);
}

/*
 * fetchm(KERN_ADDR, MY_ADDR, LENGTH) - Analogous to pfetchm, but for
 * kernal virtual memory.
 *
 */

fetchm(kern_addr, my_addr, length)
daddr_t kern_addr;
char *my_addr;
int length;
{
  int cc;

  if (kmemfd == -1) {
    kmemfd = open("/dev/kmem", O_RDONLY, 0);
    if (kmemfd < 0) {printf("Error opening /dev/kmem, errno=%d\n", errno);
		 exit(1);
	       }
  }

  if (kern_addr != kmemnextaddr)
    if (lseek(kmemfd, kern_addr, L_SET) == -1) {
      printf("Virtual lseek failed, errno=%d\n", errno);
      exit(1);
    }

  cc = read(kmemfd, my_addr, length);

  if (cc != length) {
    printf("Virtual read failed, errno=%d\n", errno);
    exit(1);
  }

  kmemnextaddr = kern_addr + length;

}

long name_lookup(name)
     char *name;
{
  struct nlist nl[2];

  nl[0].n_un.n_name = name;
  nl[1].n_un.n_name = "";

  nlist("/vmunix", nl);

  if (nl[0].n_type == 0) {
    fprintf(stderr, "Couldn't find %s, n_type=%d\n", name, nl[0].n_type);
    exit(1);
  }

  return(nl[0].n_value);
}

char * device_to_name(device)
     dev_t device;
{
#include <sys/dir.h>
#include <sys/stat.h>

  DIR *dir;
  struct direct *de;
  struct stat statbuf;
  char str[100], *newstr;

  strcpy(str, "/dev/");

  dir = opendir("/dev");

  while (de = readdir(dir)) {
    strcpy(&str[5], de->d_name);
    if (stat(str, &statbuf) == -1) {
      printf("stat died, errno=%d\n", errno);
      exit(1);
    }
    if ((statbuf.st_mode & S_IFMT) == S_IFBLK &&
	device == statbuf.st_rdev) {
      closedir(dir);
      newstr = (char *)malloc(strlen(str));
      strcpy(newstr, str);
      return(newstr);
    }
  }
  printf("Couldn't find device (%d,%d)\n", major(device), minor(device));
  exit(1);
}

/*char * read_swap(blockno)
     swblk_t blockno;
{
  static int inited = 0, nswdev, dmmax;
  static struct swdevt *swapdevs;
  static struct swapnames {
    char *name;
    int fd;
  } *swapnames;

  if (!inited) {
    int i;

    nswdev = fetch(name_lookup("_nswdev"));

    dmmax = fetch(name_lookup("_dmmax"));

    swapdevs = (struct swdevt *)calloc(nswdev, sizeof(struct swdevt));
    swapnames = (struct swapnames *)calloc(nswdev, sizeof(struct swapnames));

    fetchm(name_lookup("_swdevt"), swapdevs, nswdev * sizeof(struct swdevt));

    for (i = 0; i < nswdev; i++) {
      swapnames[i].name = device_to_name(swapdevs[i].sw_dev);
      swapnames[i].fd = open(swapnames[i].name, O_RDONLY, 0);
    }
    inited = 1;
  }

  {
    int off, seg, dev, cc;
    static char buf[DEV_BSIZE];

    off = blockno % dmmax;
    seg = blockno / dmmax;
    dev = swapnames[seg % nswdev].fd;
    seg = seg / nswdev;
    blockno = seg * dmmax + off;

    if (lseek(dev, blockno * DEV_BSIZE, L_SET) == -1) {
      printf("Swap lseek failed, errno=%d\n", errno);
      exit(1);
    }

    cc = read(dev, buf, DEV_BSIZE);

    if (cc != DEV_BSIZE) {
      printf("Swap read failed, errno=%d\n", errno);
      exit(1);
    }
    return(buf);

  }

}*/

char *
read_swap(blockno)
swblk_t blockno;
{
  static int inited = 0, swapfd;
  static char buf[DEV_BSIZE];
  static swblk_t lastblockno = 0;
  int cc;

  if (blockno == lastblockno) return(buf);

  if (!inited) {
    swapfd = open("/dev/drum", O_RDONLY, 0);
    inited = 1;
  }

  if (lseek(swapfd, blockno * DEV_BSIZE, L_SET) == -1) {
    printf("Swap lseek failed, errno=%d\n", errno);
    exit(1);
  }

  cc = read(swapfd, buf, DEV_BSIZE);

  if (cc != DEV_BSIZE) {
    printf("Swap read failed, errno=%d\n", errno);
    exit(1);
  }
  return(buf);

}

/* ufetch(p, offset) - fetch data from the u area associated with the proc
 *  table at p.
 */

long
ufetch(p, offset)
struct proc *p;
int offset;
{
  long addr;
  static int p_flag;
  static swblk_t p_swaddr;
  struct pte *pteptr;
  union {
    long pte_long;
    struct pte pte;
  } pte;
  char *buf;
  static struct proc *lastproc = 0;
  static int lastoffset = 0, lastphysaddr;

#define PGMASK (~PGOFSET)
  
  if (lastproc != p) p_flag = fetch(&p->p_flag);

  if (p_flag & SLOAD) {

    if (lastproc != p || (lastoffset & PGMASK) != (offset & PGMASK)) {
      pteptr = (struct pte *)fetch(&p->p_addr) + (offset >> PGSHIFT);
      pte.pte_long = fetch(pteptr);
      lastphysaddr = pte.pte.pg_pfnum << PGSHIFT;
      lastoffset = offset;
      lastproc = p;
    }
    addr = lastphysaddr + (offset & PGOFSET);
    return(pfetch(addr));
  } else {
    if (lastproc != p || (lastoffset & PGMASK) != (offset & PGMASK))
      p_swaddr = fetch(&p->p_swaddr);
    buf = read_swap(p_swaddr + offset / DEV_BSIZE);
    return(*(long *)(buf + offset % DEV_BSIZE));
  };
}

/* uvfetchm(procp, virtaddr, bufaddr, bytecount) - Fetch bytes from user space
 *
 * uvfetchm will fetch data from the address space of the process whose
 * proc table is addressed by procp.
 *
 */

uvfetchm(p, v, bufferaddr, bufferlength)
struct proc *p;			/* Process pointer */
caddr_t v;			/* Virtual address in user space */
unsigned char *bufferaddr;	/* Address of our buffer */
int bufferlength;		/* Number of bytes to fetch */
{
  unsigned int pageno, stkoff, szpt;
  caddr_t phyaddr;
  int count;
  struct pte *pteadr, *p0br;
  union {
    long pte_long;
    struct pte pte;
  } pte;

  while (bufferlength > 0) {
    caddr_t nextpage;
    int fragmentlength;

    nextpage = (caddr_t)((unsigned int)v + NBPG & PGMASK);
    fragmentlength = nextpage - v;

    pageno = (unsigned int)v >> PGSHIFT;

    stkoff = ((unsigned long)0x80000000 >> PGSHIFT) - pageno;

    p0br = (struct pte *)fetch(&p->p_p0br);
    szpt = (unsigned short)fetch(&p->p_szpt) * NPTEPG;
    pteadr = p0br + szpt - stkoff;
    pte.pte_long = fetch(pteadr);

    phyaddr = (caddr_t)((pte.pte.pg_pfnum << PGSHIFT)
			+ ((unsigned int)v & ~PGMASK));

    pfetchm(phyaddr, bufferaddr, fragmentlength);

    bufferaddr += fragmentlength;
    bufferlength -= fragmentlength;
    v += fragmentlength;
  }
}

struct proc *
getnextproc(p)
struct proc *p;
{

#ifndef SUN

  static struct proc *allproc = 0;

  if (!p) {
    if (!allproc) allproc = (struct proc *)name_lookup("_allproc");
    return((struct proc *)fetch(allproc));
  }

  return((struct proc *)fetch(&p->p_nxt));

#else not SUN
  
  static int inited = 0;
  static struct proc *procNPROC = 0, *CprocNPROC = 0, *proc = 0, *Cproc = 0;

  if (!inited++) {
    procNPROC = (struct proc *)name_lookup("_procNPROC");
    proc = (struct proc *)name_lookup("_proc");
    if (!procNPROC || !proc) {
      fprintf(stderr, "Can't get value of either proc or procNPROC");
      exit(1);
    }

    CprocNPROC = (struct proc *)fetch(procNPROC);
    Cproc = (struct proc *)fetch(proc);
  }

  if (!p) p = Cproc - 1;

  while (++p < CprocNPROC) {
    char p_stat; /*= (char)(fetch(&p->p_stat) >> 24);*/
      
    fetchm(&p->p_stat, &p_stat, 1);

    if(p_stat && p_stat != SZOMB) return (p);
  }

  return (0);

#endif SUN

}
