/* p - Print the process tree */

#include <sys/types.h>
#include <sys/acct.h>
#include <stdio.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/dir.h>
#include <sys/user.h>
#include <sys/proc.h>
#include <pwd.h>
#include <utmp.h>
#include <machine/vmparam.h>

time_t now;			/* GMT of now */

extern errno;

extern struct proc *getnextproc();

char *
hms(seconds, buf)
unsigned long seconds;
char *buf;
{

  if (seconds < 60 * 60) {
    sprintf(buf, "%d:%02d", seconds/60, seconds%60);
    return(buf);
  }

  sprintf(buf, "%d:%02d:%02d", seconds/3600,
	  (seconds%3600)/60, seconds%60);
    return(buf);
}

char *
username(uid)
int uid;
{
  char s[100];
  struct passwd *p;

  p = getpwuid(uid);

  if (p) return (p->pw_name);

  sprintf(s, "#%d", uid);
  return (s);
}


struct pinfo {
  struct pinfo *next;
  struct pinfo *children;
  struct pinfo *siblings;
  short pid;
  short ppid;
  short uid;
  int runtime;
  time_t starttime;
  char comm[MAXNAMLEN+1];
};

struct pinfo *plist = (struct pinfo *)0;
struct pinfo *ptree;

setup_children(parent)
struct pinfo *parent;
{
  struct pinfo *pi, *psibs = (struct pinfo *)0;

/* First, we search through plist looking for all the children of this parent */

  for(pi = plist; pi; pi = pi->next) {
    if (pi->ppid == parent->pid && pi != parent) {
      pi->siblings = parent->children;
      parent->children = pi;
    }
  }

/* Now, for each sibling, we find out it's children */

/*  for(pi = parent->children; pi; pi = pi->siblings) {
    printf("child %-10s %-6d %-6d %s\n", getpwuid(pi->uid)->pw_name,
	   pi->pid, pi->ppid, pi->comm);
  }
  printf("end of children\n");
*/

  for(pi = parent->children; pi; pi = pi->siblings) setup_children(pi);
}

print_family(parent)
struct pinfo *parent;
{
  static int indent = 0, indent_amount = 2;
  int i;

/* Print out the parent */

  for(; parent; parent = parent->siblings) {
    char buf1[25], buf2[25];

    i = indent;
    while(i--) printf(" ");

    printf("%-10s %-6d %-6d %s %s %s\n",
	   getpwuid(parent->uid)->pw_name,
	   parent->pid, parent->ppid, hms(parent->runtime, buf1),
	   hms(parent->starttime, buf2), parent->comm);

/* Now, print out the children of this parent */

    indent += indent_amount;
    print_family(parent->children);
    indent -= indent_amount;
  }
}

print_process_tree()
{
  print_family(ptree);
}

/*
 * get_process_tree will accumulate the process tree
 */

get_process_tree()
{
  struct proc *p = (struct proc *)0;
  struct pinfo *pi;
  int changeflag;

  while (p = getnextproc(p)) {
    struct user *u = (struct user *)0;
    char *cp;
    long argv, arg;

    pi = (struct pinfo *)malloc(sizeof(struct pinfo));
    pi->next = 0;
    pi->children = 0;
    pi->siblings = 0;

    if (plist == (struct pinfo *)0) plist = pi;
    else {
      pi->next = plist;
      plist = pi;
    }


/*    uid = (short)fetch(&((struct ucred *)ufetch(p, &u->u_cred))->cr_ruid);*/

    fetchm(&p->p_uid, &pi->uid, sizeof(p->p_uid));
    fetchm(&p->p_pid, &pi->pid, sizeof(p->p_pid));
    fetchm(&p->p_ppid, &pi->ppid, sizeof(p->p_ppid));
    pi->runtime = ufetch(p, &u->u_ru.ru_utime.tv_sec) +
		   ufetch(p, &u->u_ru.ru_stime.tv_sec);
#ifndef SUN
    pi->starttime = now - (time_t)ufetch(p, &u->u_start.tv_sec);
#else SUN
    pi->starttime = now - (time_t)ufetch(p, &u->u_start);
#endif

/*
    if (pi->pid > 5) {
      unsigned int buf[128], *bufp;
      unsigned int pageno = USRSTACK - sizeof(buf);

      uvfetchm(p, pageno, buf, sizeof(buf));

      printf("\nu_ap = %X\n", ufetch(p, &u->u_ap));
      for (bufp = buf+127; bufp >= buf; bufp--) {
	printf("%08X %08X %c%c%c%c\n", pageno + (bufp-buf)*4, *bufp,
	       *bufp, *bufp>>8, *bufp>>16, *bufp>>24);
      }
    }
*/
    argv = (long)u->u_comm;

    for (cp = pi->comm; cp < pi->comm + sizeof(pi->comm);
	 cp += sizeof(long), argv += sizeof(long)) {
      *(long *)cp = ufetch(p, argv);
      if (*(long *)cp == 0L) break;
    }
  }

  ptree = plist;
  do {
    if (ptree->pid == ptree->ppid) break;
    changeflag = 0;
    for (pi = plist; pi; pi = pi->next) {
      if (ptree->ppid == pi->pid) {
	ptree = pi;
	changeflag = 1;
      }
    }
  } while (changeflag);

  setup_children(ptree);

}

print_zombies()
{
  static struct proc *zombproc;
  struct proc *p;

  zombproc = (struct proc *)name_lookup("_zombproc");

  if (zombproc == (struct proc *)0) {
    printf("Can't find definition of zombproc\n");
    exit(1);
  }

  p = (struct proc *)fetch(zombproc);
  if (p == (struct proc *)0) {
    printf("No zombies\n");
    return;
  }

  do
    printf("%s %d %d\n", getpwuid((short)fetch(&p->p_uid))->pw_name,
	   (short)fetch(&p->p_pid),
	   (short)fetch(&p->p_ppid));
  while (p = getnextproc(p));
}

main()
{

#ifndef SUN
  extern int _pw_stayopen;

  _pw_stayopen = 1;		/* Speeds up passwd db refs */
#endif not SUN

  now = (time_t)time((long *)0);

  get_process_tree();
  print_process_tree();
#ifndef SUN
  print_zombies();
#endif
}
